//---------------------------------------------------------------------------
// build.cpp
// part of PVBUILD
// (c) 2006 by Moritz Beutel, AUDACIA Software
//---------------------------------------------------------------------------


#include "build.hpp"
#include "files.hpp"

#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <cctype>
#include <windows.h>
#include "ucl/file.hpp"

using namespace std;


std::string intToStr (int val, int radix = 10)
{
  char buf[33];
  return (itoa (val, buf, radix));
}

std::string getExePath (void)
{
  char str[MAX_PATH];
  GetModuleFileName (NULL, str, MAX_PATH);
  return (ucl::extractFilePath (str));
}

BuildTool::BuildTool (const ucl::IniFile& set, BuildInfo& _bi)
  : settings (set), bi (_bi)
{
  bi.memmodel = toupper (bi.memmodel);
  if ((bi.memmodel != 'S') && ((bi.memmodel != 'P')))
    bi.memmodel = 'S';
  exepath = getExePath ();
}

bool BuildTool::findCbxFile (string& filename)
{
  WIN32_FIND_DATA fd;
  HANDLE hFind;

  hFind = FindFirstFile ((filename + ".cbx").c_str (), &fd);
  if (hFind == INVALID_HANDLE_VALUE)
    hFind = FindFirstFile ("*.cbx", &fd);
  if (hFind == INVALID_HANDLE_VALUE)
  {
    cerr << "Unable to find project file" << endl;
    filename.clear ();
    return (false);
  }

  filename = ucl::extractFileName (fd.cFileName);
  return (true);
}

bool BuildTool::parseXmlNode (TiXmlNode* node)
{
  TiXmlNode* subnode;
  TiXmlElement* element;
  const char* fn;

  if (!node)
    return (false);
  subnode = node->FirstChild ();
  while (subnode)
  {
    if (subnode->Type () == TiXmlNode::ELEMENT)
    {
      if (subnode->ValueStr () == "node")
        parseXmlNode (subnode);
      if (subnode->ValueStr () == "file")
      {
        element = subnode->ToElement ();
        if (element)
        {
          fn = element->Attribute ("path");
          if (fn)
            files.push_back (fn);
        }
      }
    }
    subnode = subnode->NextSibling ();
  }
  return (true);
}

bool BuildTool::WriteMakefile ()
{
  FILE* file;
  TiXmlDocument xml;
  TiXmlNode* node;
  TiXmlNode* node2;
  string iconfile, liconfile;
  string ext, name, path;
  int n;

  bi.filename = ucl::extractFileName (bi.filename);
  if (!findCbxFile (bi.filename))
    return (false);
  if (bi.progname.empty ())
    bi.progname = bi.filename;


  xml.LoadFile (bi.filename + ".cbx", TIXML_ENCODING_UNKNOWN);
  node = xml.FirstChild ("project");
  if (!node)
  {
    cerr << "Error: Wrong format in .cbx file" << endl;
    return (false);
  }

  if (!parseXmlNode (node))
  {
    cerr << "Error: Wrong format in .cbx file" << endl;
    return (false);
  }

  for (int i = files.size () - 1; i >= 0; --i)
  {
    for (n = files[i].size () - 1; n >= 0; --n)
      if (files[i][n] == '/')
        files[i][n] = '\\';

    ext  = ucl::extractFileExt (files[i]);
    for (n = ext.size () - 1; n >= 0; --n)
      ext[n] = tolower (ext[n]);

    if ((ext == ".c") || (ext == ".a86") || (ext == ".asm"))
      addObjFile (files[i]);
    if (ext == ".h")
      addHdrFile (files[i]);
    if (ext == ".lib")
      addLibFile (files[i]);
  }


  intWriteMakefile ();

  return (true);
}



PV2BuildTool::PV2BuildTool (const ucl::IniFile& set, BuildInfo& _bi)
  : BuildTool (set, _bi)
{
  sdkpath = settings["PV2SDK Paths"]["SDKPath"];
}


void PV2BuildTool::addObjFile (const std::string& objfile)
{
  string path = ucl::extractFilePath (objfile);

  objfiles += "$(ODIR)\\" + ucl::extractFileName (objfile) + ".obj \\\n\t";
}

void PV2BuildTool::addHdrFile (const std::string& hdrfile)
{
  string path = ucl::extractFilePath (hdrfile);

  hdrfiles += ((ucl::toLowerCase (path) == "h\\") ? string ("$(HDIR)\\") : path)
    + ucl::extractFileName (hdrfile) + ".h \\\n\t";
}

void PV2BuildTool::addLibFile (const std::string& libfile)
{
  libfiles += " $(CLDIR)\\" + ucl::extractFileNameExt (libfile);
}


bool PV2BuildTool::intWriteMakefile ()
{
  FILE* file;
  string buf;
  static const char dbgpath_s[] = "PV2BUILD\\OBJDBG_S";
  static const char relpath_s[] = "PV2BUILD\\OBJREL_S";
  static const char dbgpath_p[] = "PV2BUILD\\OBJDBG_P";
  static const char relpath_p[] = "PV2BUILD\\OBJREL_P";
  static const char dbgpath[]   = "PV2BUILD\\OBJDBG";
  static const char relpath[]   = "PV2BUILD\\OBJREL";

  const char* objdir = bi.library ? "_$(MODEL)" : "";


    // create directories
  CreateDirectory ("PV2BUILD", NULL);
  if (bi.library)
  {
    CreateDirectory (dbgpath_s, NULL);
    CreateDirectory (relpath_s, NULL);
    CreateDirectory (dbgpath_p, NULL);
    CreateDirectory (relpath_p, NULL);
  }
  else
  {
    CreateDirectory (dbgpath, NULL);
    CreateDirectory (relpath, NULL);
  }


    /* write MakeSDK.1 */
  file = fopen ("PV2BUILD\\MakeSDK.1", "wt");
  if (!file)
  {
    cerr << "Unable to write PV2BUILD\\MakeSDK.1" << endl;
    return (false);
  }

  buf.clear ();
  for (int i = bi.defines.size () - 1; i >= 0; --i)
    buf += "-D" + bi.defines[i] + " ";

  fprintf (file, pv2_res_makeSDK1, objdir, objdir, sdkpath.c_str (), (bi.asmopt ? "ASMOPT =" : ""),
           buf.c_str ());

  fclose (file);


    /* write MakeSDK.2 */
  file = fopen ("PV2BUILD\\MakeSDK.2", "wt");
  if (!file)
  {
    cerr << "Unable to write PV2BUILD\\MakeSDK.2" << endl;
    return (false);
  }

  fprintf (file, pv2_res_makeSDK2, bi.tarseg, libfiles.c_str ());

  fclose (file);


    /* write makefile */
  file = fopen ("PV2BUILD\\makefile", "wt");
  if (!file)
  {
    cerr << "Unable to write PV2BUILD\\makefile" << endl;
    return (false);
  }

    /* version, version, targetname, progname,
       objfiles, hdrfiles, libadd */
  fprintf (file, pv2_res_makefile, bi.version, bi.version, bi.progname.c_str (),
           bi.progname.c_str (), objfiles.c_str (), hdrfiles.c_str (),
           bi.library ? pv2_res_makefile_libadd : "");

  fclose (file);

  return (true);
}

bool PV2BuildTool::invokeMakeTool (char model)
{
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  BOOL success;
  char params[MAX_PATH];


    /* call kmmake */
  ZeroMemory (&si, sizeof (STARTUPINFO));
  si.cb = sizeof (STARTUPINFO);

  strcpy (params, "kmmake -f PV2BUILD\\makefile ");
  if (bi.library | bi.debug)
    strcat (params, "-i ");
  if (bi.debug)
    strcat (params, "DEBUG=1 ");
  if (model == 'P')
    strcat (params, "PMODEL=1 ");

  success = CreateProcess (NULL, params, NULL, NULL, true,
    CREATE_DEFAULT_ERROR_MODE | NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi);

  WaitForSingleObject (pi.hProcess, INFINITE);

  DeleteFile ("make.i");

  if (success)
  {
    CloseHandle (pi.hProcess);
    CloseHandle (pi.hThread);
  }
  else
    cerr << "Error: Unable to start kmmake" << endl;

  return (success);
}

bool PV2BuildTool::DoBuild ()
{
  char* path;
  char* curpath;
  string binpath = settings["PV2SDK Paths"]["CompilerPath"].asString () + "\\bin;";



    /* add paths */
  curpath = getenv ("PATH");
  path = (char*) malloc (strlen (curpath) + 5 /* "PATH=" */ + sdkpath.size ()
                       + 6 /* "\\TOOL;" */ + binpath.size () + 2 /* \0\0 */);
  strcpy (path, "PATH=");
  strcat (path, (sdkpath + "\\TOOL;").c_str ());
  strcat (path, binpath.c_str ());
  strcat (path, curpath);

  putenv (path);

  free (path);


    /* TODO: add library to LINK2$(MODEL).DAT */
  return (bi.library
        ? (invokeMakeTool ('P') && invokeMakeTool ('S'))
        : invokeMakeTool (bi.memmodel));
}







PV3BuildTool::PV3BuildTool (const ucl::IniFile& set, BuildInfo& _bi)
  : BuildTool (set, _bi), objcount (0)
{
  sdkpath = settings["PV3SDK Paths"]["SDKPath"];
}


void PV3BuildTool::addObjFile (const std::string& objfile)
{
  if (ucl::extractFileExt (objfile) == ".c") // C files only
    objfiles += "FILE" + intToStr (objcount++) + "=" + ucl::extractFileName (objfile) + "\n";
}

void PV3BuildTool::addHdrFile (const std::string&)
{
  /* not needed for PV3SDK */
}

void PV3BuildTool::addLibFile (const std::string& libfile)
{
  string filename = ucl::extractFileNameExt (libfile);
  libfiles += " $(PVSDK_BASE)\\lib\\" + filename;
  libinc += "library $(PVSDK_BASE)\\lib\\" + filename + "\n";
}



bool PV3BuildTool::intWriteMakefile ()
{
  FILE* file;
  string buf;
  static const char dbgpath[] = "PV3BUILD\\OBJDBG";
  static const char relpath[] = "PV3BUILD\\OBJREL";
  int i;
  string binpath = settings["PV3SDK Paths"]["CompilerPath"].asString ();
  string filemacro;
  int objcnt = objcount;


  objcount = 0;

    // create directories
  CreateDirectory ("PV3BUILD", NULL);
  CreateDirectory (dbgpath, NULL);
  CreateDirectory (relpath, NULL);



  buf.reserve ((i = bi.defines.size ()) * 30);
  for (--i; i >= 0; --i)
    buf += "-define=" + bi.defines[i] + " ";

    /* write makefile */
  file = fopen ("PV3BUILD\\makefile.mak", "wt");
  if (!file)
  {
    cerr << "Unable to write PV3BUILD\\makefile.mak" << endl;
    return (false);
  }

     /* binpath, sdkpath, subfile, outdir, libraries, dbgrel */
  fprintf (file, pv3_res_makefile1, binpath.c_str (), sdkpath.c_str (),
           bi.debug ? "MakeDbg.sub" : "Makefile.sub", bi.debug ? dbgpath : relpath,
           libfiles.c_str (), (bi.debug || bi.library) ? "DBG" : "REL");

     /* targetname, progname, version, defines, objfiles */
  fprintf (file, pv3_res_sourcesdef, bi.progname.c_str (), bi.progname.c_str (),
           bi.version, buf.c_str (), objfiles.c_str ());

  fputs ("USERALLOBJ = \\\n", file);

  for (i = 0; i < objcnt; ++i)
    fputs (("\t\"$(OUTDIR)\\$(FILE" + intToStr (i) + ").obj\" \\\n").c_str (), file);
  fputs ("\n\n!MESSAGE --- Executing SH C/C++ Compiler phase ---\n", file);

  filemacro.reserve (12);
  for (i = 0; i < objcnt; ++i)
  {
    filemacro = "$(FILE" + intToStr (i) + ")";
    fputs (("\"$(OUTDIR)\\" + filemacro + ".obj\" : \"$(SRCDIR)\\" + filemacro + ".C\" $(LNKINF0)\n"
            "\t$(CC) \"$(SRCDIR)\\" + filemacro + ".c\" -objectfile=\"$(OUTDIR)\\" + filemacro
            + ".obj\" $(CCINF0)\n\n").c_str (), file);
  }

    /* pvalib, rldlib */
  fprintf (file, pv3_res_makefile2, bi.library ? "lib" : (bi.debug ? "abs" : "pva"),
           bi.library ? "lib" : (bi.debug ? "abs" : "rld"));
  if (bi.library)
    fputs (pv3_res_makefile3_lib, file);
  else
    fprintf (file, bi.debug ? pv3_res_makefile3_dbg : pv3_res_makefile3_rel, libinc.c_str ());

  fclose (file);


    /* write sources.def */
  file = fopen ("PV3BUILD\\sources.def", "wt");
  if (!file)
  {
    cerr << "Unable to write PV3BUILD\\sources.def" << endl;
    return (false);
  }

       /* targetname, progname, version, defines, objfiles */
  fprintf (file, pv3_res_sourcesdef, bi.progname.c_str (), bi.progname.c_str (),
           bi.version, buf.c_str (), objfiles.c_str ());

  fclose (file);

  return (true);
}




bool PV3BuildTool::DoBuild ()
{
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  BOOL success;
  char params[MAX_PATH];

  char* path;
  char* curpath;
  string binpath = settings["PV3SDK Paths"]["CompilerPath"].asString ();



    /* add paths */
  curpath = getenv ("PATH");
  path = (char*) malloc (strlen (curpath) + 5 /* "PATH=" */ + sdkpath.size ()
                       + 7 /* "\\TOOLS;" */ + binpath.size () + 2 /* \0\0 */);
  strcpy (path, "PATH=");
  strcat (path, (sdkpath + "\\TOOLS;").c_str ());
  strcat (path, binpath.c_str ());
  strcat (path, curpath);

  putenv (path);
  free (path);

  putenv (("SHC_INC=" + binpath + "\\include").c_str ());
  putenv (("SHC_LIB=" + binpath + "\\bin").c_str ());
  putenv (("SHC_TMP=" + binpath).c_str ());


    /* call hmake */
  ZeroMemory (&si, sizeof (STARTUPINFO));
  si.cb = sizeof (STARTUPINFO);

  strcpy (params, binpath.c_str ());
  strcat (params, "\\bin\\hmake PV3BUILD\\makefile.mak");

  success = CreateProcess (NULL, params, NULL, NULL, true,
    CREATE_DEFAULT_ERROR_MODE | NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi);

  WaitForSingleObject (pi.hProcess, INFINITE);

  if (success)
  {
    CloseHandle (pi.hProcess);
    CloseHandle (pi.hThread);
  }
  else
    cerr << "Error: Unable to start hmake" << endl;

  return (success);
}


